@@ -36,6 +36,7 @@ struct Imports {
 #[derive(Default)]
 struct Exports {
     freestanding_funcs: Vec<Source>,
+    arg_types: Vec<Source>,
     resource_funcs: BTreeMap<ResourceId, Vec<Source>>,
 }
 
@@ -66,19 +67,6 @@ impl Ts {
         }
     }
 
-    fn is_nullable_option(&self, iface: &Interface, variant: &Variant) -> bool {
-        match variant.as_option() {
-            Some(ty) => match ty {
-                Type::Id(id) => match &iface.types[*id].kind {
-                    TypeDefKind::Variant(v) => !self.is_nullable_option(iface, v),
-                    _ => true,
-                },
-                _ => true,
-            },
-            None => false,
-        }
-    }
-
     fn array_ty(&self, iface: &Interface, ty: &Type) -> Option<&'static str> {
         match ty {
             Type::U8 | Type::CChar => Some("Uint8Array"),
@@ -100,76 +88,20 @@ impl Ts {
         }
     }
 
-    fn ty_to_str(&self, iface: &Interface, ty: &Type) -> String {
-        match ty {
-            Type::U8
-            | Type::CChar
-            | Type::S8
-            | Type::U16
-            | Type::S16
-            | Type::U32
-            | Type::Usize
-            | Type::S32
-            | Type::F32
-            | Type::F64 => "number".to_string(),
-            Type::U64  => "u64".to_string(),
-            Type::S64 => "i64".to_string(),
-            Type::Char => "string".to_string(),
-            Type::Handle(id) => iface.resources[*id].name.to_camel_case(),
-            Type::Id(id) => {
-                let ty = &iface.types[*id];
-                if let Some(name) = &ty.name {
-                    return name.to_camel_case();
-                }
-                match &ty.kind {
-                    TypeDefKind::Type(t) => self.ty_to_str(iface, t),
-                    TypeDefKind::Record(r) if r.is_tuple() => self.tuple_to_str(iface, r),
-                    TypeDefKind::Record(_) => panic!("anonymous record"),
-                    TypeDefKind::Variant(v) if v.is_bool() => "boolean".to_string(),
-                    TypeDefKind::Variant(v) => {
-                        if self.is_nullable_option(iface, v) {
-                            return self.ty_to_str(iface, v.cases[1].ty.as_ref().unwrap())
-                                + " | null";
-                        } else if let Some(t) = v.as_option() {
-                            // self.needs_ty_option = true;
-                            return format!("Option<{}>", self.ty_to_str(iface, t));
-                        } else if let Some((ok, err)) = v.as_expected() {
-                            // self.needs_ty_result = true;
-                            let first = match ok {
-                                Some(ok) => self.ty_to_str(iface, ok),
-                                None => "undefined".to_string(),
-                            };
-                            let second = match err {
-                                Some(err) => self.ty_to_str(iface, err),
-                                None => "undefined".to_string(),
-                            };
-                            return format!("Result<{}, {}>", first, second);
-                        }
-                        panic!("anonymous variant");
-                    }
-                    TypeDefKind::List(v) => self.list_to_str(iface, v),
-                    TypeDefKind::PushBuffer(_) => "buffer".to_string(), //self.print_buffer(iface, true, _),
-                    TypeDefKind::PullBuffer(_) => "buffer".to_string(), //self.print_buffer(iface, false, _),
-                    TypeDefKind::Pointer(_) | TypeDefKind::ConstPointer(_) => "number".to_string(),
-                }
-            }
-        }
-    }
-
     fn print_ty(&mut self, iface: &Interface, ty: &Type) {
         match ty {
-            Type::U8
-            | Type::CChar
-            | Type::S8
-            | Type::U16
-            | Type::S16
-            | Type::U32
-            | Type::Usize
-            | Type::S32
-            | Type::F32
-            | Type::F64 => self.src.ts("number"),
-            Type::U64  => self.src.ts("u64"),
+            Type::U8 => self.src.ts("u8"),
+            Type::S8 => self.src.ts("i8"),
+            Type::U16 => self.src.ts("u16"),
+            Type::S16 => self.src.ts("i16"),
+            Type::U32 => self.src.ts("u32"),
+            Type::Usize => self.src.ts("usize"),
+            Type::S32 => self.src.ts("i32"),
+            Type::F32 => self.src.ts("f32"),
+            Type::F64 => self.src.ts("f64"),
+            Type::U64 => self.src.ts("u64"),
             Type::S64 => self.src.ts("i64"),
+            Type::CChar => self.src.ts("number"),
             Type::Char => self.src.ts("string"),
             Type::Handle(id) => self.src.ts(&iface.resources[*id].name.to_camel_case()),
             Type::Id(id) => {
@@ -183,7 +115,7 @@ impl Ts {
                     TypeDefKind::Record(_) => panic!("anonymous record"),
                     TypeDefKind::Variant(v) if v.is_bool() => self.src.ts("boolean"),
                     TypeDefKind::Variant(v) => {
-                        if self.is_nullable_option(iface, v) {
+                        if iface.is_nullable_option(v) {
                             self.print_ty(iface, v.cases[1].ty.as_ref().unwrap());
                             self.src.ts(" | null");
                         } else if let Some(t) = v.as_option() {
@@ -219,26 +151,13 @@ impl Ts {
         }
     }
 
-    fn list_to_str(&self, iface: &Interface, ty: &Type) -> String {
-        if let Some(src) = self.hash_map_to_str(iface, ty) {
-            src
-        } else {
-            match self.array_ty(iface, ty) {
-                Some(ty) => ty.to_string(),
-                None => {
-                    if let Type::Char = ty {
-                        "string".to_string()
-                    } else {
-                        format!("{}[]", self.ty_to_str(iface, ty))
-                    }
-                }
-            }
-        }
-    }
-
     fn print_list(&mut self, iface: &Interface, ty: &Type) {
-        if let Some(src) = self.hash_map_to_str(iface, ty) {
-            self.src.ts(&src)
+        if let Some(r) = self.hash_map(iface, ty) {
+            self.src.ts("Record<");
+            self.print_ty(iface, &r.fields[0].ty);
+            self.src.ts(", ");
+            self.print_ty(iface, &r.fields[1].ty);
+            self.src.ts(">");
         } else {
             match self.array_ty(iface, ty) {
                 Some(ty) => self.src.ts(ty),
@@ -254,18 +173,6 @@ impl Ts {
         }
     }
 
-    fn tuple_to_str(&self, iface: &Interface, record: &Record) -> String {
-        format!(
-            "[{}]",
-            record
-                .fields
-                .iter()
-                .map(|field| self.ty_to_str(iface, &field.ty))
-                .collect::<Vec<_>>()
-                .join(", ")
-        )
-    }
-
     fn print_tuple(&mut self, iface: &Interface, record: &Record) {
         self.src.ts("[");
         for (i, field) in record.fields.iter().enumerate() {
@@ -296,7 +203,11 @@ impl Ts {
     }
 
     fn docs(&mut self, docs: &Docs) {
-        let docs = match &docs.contents {
+        self.doc_str(&docs.contents);
+    }
+
+    fn doc_str(&mut self, contents: &Option<String>) {
+        let docs = match &contents {
             Some(docs) => docs,
             None => return,
         };
@@ -304,7 +215,7 @@ impl Ts {
             .lines()
             .filter(|line| *line != "change" && *line != "view")
             .collect::<Vec<&str>>();
-        if lines.len() > 0 {
+        if !lines.is_empty() {
             self.src.ts("/**\n");
             for line in lines {
                 self.src.ts(&format!("* {}\n", line));
@@ -313,6 +224,34 @@ impl Ts {
         }
     }
 
+    fn print_args_type(&mut self, iface: &Interface, func: &Function, param_start: usize) {
+        let none: Option<String> = None;
+        let arg_fields: Vec<(&str, &Type, &Option<String>)> = func.params[param_start..]
+            .iter()
+            .map(|(name, ty)| (name.as_str(), ty, &none))
+            .collect();
+        self.print_fields(iface, arg_fields);
+    }
+
+    fn print_args(&mut self, iface: &Interface, func: &Function, param_start: usize) {
+        self.src.ts("(args");
+        if func.params.is_empty() {
+            self.src.ts(" = {}");
+        } else {
+            self.src.ts(": {\n");
+            self.print_args_type(iface, func, param_start);
+            self.src.ts("}");
+        }
+        if func.params.is_empty() {};
+        self.src.ts(", options?: ");
+        self.src.ts(if is_change(func) {
+            "ChangeMethodOptions"
+        } else {
+            "ViewFunctionOptions"
+        });
+        self.src.ts("): ");
+    }
+
     fn ts_func(&mut self, iface: &Interface, func: &Function) {
         self.docs(&func.docs);
         if is_change(func) {
@@ -343,7 +282,7 @@ impl Ts {
                 self.src.ts(&iface.name.to_mixed_case());
                 self.src.ts(": ");
                 self.src.ts(&iface.name.to_camel_case());
-                if func.params.len() > 0 {
+                if !func.params.is_empty() {
                     self.src.ts(", ");
                 }
                 0
@@ -353,34 +292,36 @@ impl Ts {
         };
         let name = func.item_name().to_snake_case();
 
-        let mut args_string = String::new();
-
-        for (i, (name, ty)) in func.params[param_start..].iter().enumerate() {
-            if i > 0 {
-                args_string.push_str(", ");
-            }
-            let (_type, is_nullable) = is_nullable(iface, ty);
-            args_string.push_str(to_js_ident(&name.to_snake_case()));
-            if is_nullable {
-                args_string.push_str("?");
-            }
-            args_string.push_str(": ");
-            args_string.push_str(&self.ty_to_str(iface, &_type));
-        }
-        let default_object = if args_string.len() > 0 { "" } else { " = {}" };
-        let options_type = if is_change(func) {
-            "ChangeMethodOptions"
-        } else {
-            "ViewFunctionOptions"
-        };
-        let arg_str =
-            format!("(args: {{{args_string}}}{default_object}, options?: {options_type}): ");
-
-        self.src.ts(&arg_str);
+        self.print_args(iface, func, param_start);
 
         // Always async
         self.src.ts("Promise<");
 
+        self.print_func_result(iface, func);
+
+        self.src.ts("> {\n");
+
+        if is_change(func) {
+            self.src.ts(&format!(
+                "return providers.getTransactionLastResult(await this.{name}Raw(args, options));\n}}\n"
+            ));
+            self.docs(&func.docs);
+            self.src.ts(&format!("{name}Raw"));
+            self.print_args(iface, func, param_start);
+            self.src.ts("Promise<providers.FinalExecutionOutcome> {\n");
+            self.src.ts(&format!("return this.account.functionCall({{contractId: this.contractId, methodName: \"{name}\", args, ...options}});\n}}\n"));
+            self.docs(&func.docs);
+            self.src.ts(&format!("{name}Tx"));
+            self.print_args(iface, func, param_start);
+            self.src.ts(&format!("transactions.Action {{\n return transactions.functionCall(\"{name}\", args, options?.gas ?? DEFAULT_FUNCTION_CALL_GAS, options?.attachedDeposit ?? new BN(0))\n}}\n"));
+        } else {
+            self.src.ts(&format!(
+                "return this.account.viewFunction(this.contractId, \"{name}\", args, options);\n}}\n"
+            ));
+        }
+    }
+
+    fn print_func_result(&mut self, iface: &Interface, func: &Function) {
         match func.results.len() {
             0 => self.src.ts("void"),
             1 => self.print_ty(iface, &func.results[0].1),
@@ -408,25 +349,19 @@ impl Ts {
                 }
             }
         }
+    }
 
-        self.src.ts("> {\n");
-
-        if is_change(func) {
-            self.src.ts(&format!(
-                "return providers.getTransactionLastResult(await this.{name}Raw(args, options));\n}}\n"
-            ));
-            self.docs(&func.docs);
-            self.src.ts(&format!(
-                "{name}Raw{arg_str} Promise<providers.FinalExecutionOutcome> {{\n"
-            ));
-            self.src.ts(&format!("return this.account.functionCall({{contractId: this.contractId, methodName: \"{name}\", args, ...options}});\n}}\n"));
-            self.docs(&func.docs);
-            self.src
-                .ts(&format!("{name}Tx{arg_str} transactions.Action {{\n return transactions.functionCall(\"{name}\", args, options?.gas ?? DEFAULT_FUNCTION_CALL_GAS, options?.attachedDeposit ?? new BN(0))\n}}\n"));
-        } else {
-            self.src.ts(&format!(
-                "return this.account.viewFunction(this.contractId, \"{name}\", args, options);\n}}\n"
-            ));
+    fn print_fields(&mut self, iface: &Interface, fields: Vec<(&str, &Type, &Option<String>)>) {
+        for (name, ty, docs) in fields.iter() {
+            self.doc_str(docs);
+            self.src.ts(&name.to_snake_case());
+            if iface.is_ty_nullable_option(ty) {
+                self.src.ts("?");
+            }
+            self.src.ts(": ");
+            let ty = iface.get_nullable_option(ty).unwrap_or(ty);
+            self.print_ty(iface, ty);
+            self.src.ts(";\n");
         }
     }
 }
@@ -436,26 +371,7 @@ impl Generator for Ts {
         let variant = Self::abi_variant(dir);
         self.sizes.fill(variant, iface);
         self.in_import = variant == AbiVariant::GuestImport;
-        self.src
-            .ts("import { Account, transactions, providers, DEFAULT_FUNCTION_CALL_GAS } from 'near-api-js';\n\n");
-        self.src.ts("
-        import BN from 'bn.js';
-        export interface ChangeMethodOptions {
-          gas?: BN;
-          attachedDeposit?: BN;
-          walletMeta?: string;
-          walletCallbackUrl?: string;
-      }
-      export interface ViewFunctionOptions {
-        parse?: (response: Uint8Array) => any;
-        stringify?: (input: any) => any;
-      }
-
-      /** 64 bit unsigned integer less than 2^53 -1 */
-      type u64 = number;
-      /** 64 bit signed integer less than 2^53 -1 */
-      type i64 = number;
-        ")
+        self.add_preamble()
     }
 
     fn type_record(
@@ -499,17 +415,13 @@ impl Generator for Ts {
         } else {
             self.src
                 .ts(&format!("export interface {} {{\n", name.to_camel_case()));
-            for field in record.fields.iter() {
-                self.docs(&field.docs);
-                let (_ty, is_nullable) = is_nullable(iface, &field.ty);
-                self.src.ts(&format!(
-                    "{}{}: ",
-                    field.name.to_snake_case(),
-                    if is_nullable { "?" } else { "" }
-                ));
-                self.print_ty(iface, &_ty);
-                self.src.ts(",\n");
-            }
+
+            let fields = record
+                .fields
+                .iter()
+                .map(|f| (f.name.as_str(), &f.ty, &f.docs.contents))
+                .collect();
+            self.print_fields(iface, fields);
             self.src.ts("}\n");
         }
     }
@@ -528,7 +440,7 @@ impl Generator for Ts {
                 "export type {} = boolean;\n",
                 name.to_camel_case(),
             ));
-        } else if self.is_nullable_option(iface, variant) {
+        } else if iface.is_nullable_option(variant) {
             self.src
                 .ts(&format!("export type {} = ", name.to_camel_case()));
             self.print_ty(iface, variant.cases[1].ty.as_ref().unwrap());
@@ -536,10 +448,10 @@ impl Generator for Ts {
         } else if variant.is_enum() {
             self.src
                 .ts(&format!("export enum {} {{\n", name.to_camel_case()));
-            for (i, case) in variant.cases.iter().enumerate() {
+            for case in variant.cases.iter() {
                 self.docs(&case.docs);
                 let name = case.name.to_camel_case();
-                self.src.ts(&format!("{} = {},\n", name, i));
+                self.src.ts(&format!("{} = \"{}\",\n", name, name));
             }
             self.src.ts("}\n");
         } else {
@@ -603,10 +515,12 @@ impl Generator for Ts {
         ty: &Type,
         docs: &Docs,
     ) {
+        #[allow(clippy::drop_copy)]
         drop((iface, _id, name, const_, ty, docs));
     }
 
     fn type_builtin(&mut self, iface: &Interface, _id: TypeId, name: &str, ty: &Type, docs: &Docs) {
+        #[allow(clippy::drop_copy)]
         drop((iface, _id, name, ty, docs));
     }
 
@@ -667,10 +581,56 @@ impl Generator for Ts {
                 exports
                     .resource_funcs
                     .entry(*resource)
-                    .or_insert(Vec::new())
+                    .or_insert_with(Vec::new)
                     .push(func_body);
             }
+        };
+        let prev = mem::take(&mut self.src);
+        let is_change_func = is_change(func);
+        let func_type = if is_change_func { "change" } else { "view" };
+        let docs = func
+            .docs
+            .contents
+            .as_ref()
+            .map(|d| d.to_string())
+            .unwrap_or_default();
+        self.doc_str(&Some(format!("{docs}\n@contractMethod {func_type}\n")));
+        self.src.ts(&format!(
+            "export interface {} {{\n",
+            func.name.to_camel_case()
+        ));
+
+        self.src.ts("args: {");
+        if !func.params.is_empty() {
+            self.src.ts("\n");
+            self.print_args_type(iface, func, 0);
         }
+        self.src.ts("};\n");
+        if is_change_func {
+            self.src.ts("options: {
+      /** Units in gas
+       * @pattern [0-9]+
+       * @default \"30000000000000\"
+       */
+      gas?: string;
+      /** Units in yoctoNear
+       * @default 0
+       */
+      attachedDeposit?: Balance;
+    }\n");
+        }
+        self.src.ts("\n}\n");
+        self.src.ts(&format!("export type {}__Result = ", func.name.to_camel_case()));
+        self.print_func_result(iface, func);
+        self.src.ts(";\n");
+
+        let func_args = mem::replace(&mut self.src, prev);
+        let exports = self
+            .guest_exports
+            .entry(iface.name.to_string())
+            .or_insert_with(Exports::default);
+
+        exports.arg_types.push(func_args);
     }
 
     fn finish_one(&mut self, iface: &Interface, files: &mut Files) {
@@ -698,7 +658,6 @@ impl Generator for Ts {
             }
         }
         let imports = mem::take(&mut self.src);
-
         for (_module, exports) in mem::take(&mut self.guest_exports) {
             self.src.ts("\nexport class Contract {
                   
@@ -707,10 +666,11 @@ impl Generator for Ts {
                 self.src.ts(&func.ts);
             }
             self.src.ts("}\n");
+            for args in exports.arg_types.iter() {
+                self.src.ts(&args.ts);
+            }
         }
 
-        let exports = mem::take(&mut self.src);
-
         if mem::take(&mut self.needs_ty_option) {
             self.src
                 .ts("export type Option<T> = { tag: \"none\" } | { tag: \"some\", val; T };\n");
@@ -720,15 +680,15 @@ impl Generator for Ts {
                 "export type Result<T, E> = { tag: \"ok\", val: T } | { tag: \"err\", val: E };\n",
             );
         }
+        let exports = mem::take(&mut self.src);
 
         self.src.ts(&imports.ts);
         self.src.ts(&exports.ts);
 
         let src = mem::take(&mut self.src);
         let name = iface.name.to_kebab_case();
-        if !self.opts.no_typescript {
-            files.push(&format!("{}.ts", name), src.ts.as_bytes());
-        }
+        files.push(&format!("{}.ts", name), src.ts.as_bytes());
+        files.push("helper.ts", HELPER.as_bytes());
     }
 
     fn finish_all(&mut self, _files: &mut Files) {
@@ -737,24 +697,31 @@ impl Generator for Ts {
 }
 
 impl Ts {
-    fn hash_map_to_str(&self, iface: &Interface, ty: &Type) -> Option<String> {
-        match ty {
-            Type::Id(id) => {
-                let ty = &iface.types[*id];
-                if let Some(_) = &ty.name {
-                    return None;
-                }
-                match &ty.kind {
-                    TypeDefKind::Record(r) if r.is_tuple() && r.fields.len() == 2 => Some(format!(
-                        "Record<{}, {}>",
-                        self.ty_to_str(iface, &r.fields[0].ty),
-                        self.ty_to_str(iface, &r.fields[1].ty)
-                    )),
-                    _ => None,
-                }
-            }
-            _ => None,
-        }
+    fn hash_map<'a>(&self, iface: &'a Interface, ty: &Type) -> Option<&'a Record> {
+        iface
+            .get_record(ty)
+            .filter(|r| r.is_tuple() && r.fields.len() == 2)
+    }
+    fn add_preamble(&mut self) {
+        self.src.ts("import {
+  Account,
+  transactions,
+  providers,
+  DEFAULT_FUNCTION_CALL_GAS,
+  u8,
+  i8,
+  u16,
+  i16,
+  u32,
+  i32,
+  u64,
+  i64,
+  f32,
+  f64,
+  BN,
+  ChangeMethodOptions,
+  ViewFunctionOptions,
+} from './helper';\n\n");
     }
 }
 
@@ -779,29 +746,99 @@ impl Source {
 
 fn is_change(func: &Function) -> bool {
     if let Some(docs) = &func.docs.contents {
-        let x = docs
-            .split("\n")
-            .filter(|s| *s == "change")
-            .collect::<Vec<_>>();
-        if x.len() == 1 {
+        let mut x = docs.split('\n').filter(|s| *s == "change").peekable();
+        if x.peek().is_some() {
             return true;
         }
     }
     false
 }
 
-// TODO replace this with work upstream
-fn is_nullable(iface: &Interface, ty: &Type) -> (Type, bool) {
-    //Note: currently making type non-nullable since the "?" makes it optional
-    if let Type::Id(id) = ty {
-        match &iface.types[*id].kind {
-            TypeDefKind::Variant(v) => v
-                .as_option()
-                .and_then(|_| v.cases[1].ty.as_ref())
-                .map_or((ty.clone(), false), |ty| (ty.clone(), true)),
-            _ => (ty.clone(), false),
-        }
-    } else {
-        (ty.clone(), false)
-    }
+const HELPER: &str = "
+//@ts-ignore for ts-json-schema-generator
+export { Account, transactions, providers, DEFAULT_FUNCTION_CALL_GAS } from 'near-api-js';
+//@ts-ignore for ts-json-schema-generator
+import BN from 'bn.js';
+export {BN};
+
+export interface ChangeMethodOptions {
+  gas?: BN;
+  attachedDeposit?: BN;
+  walletMeta?: string;
+  walletCallbackUrl?: string;
 }
+/**
+ * Options for view contract calls
+ */ 
+export interface ViewFunctionOptions {
+  parse?: (response: Uint8Array) => any;
+  stringify?: (input: any) => any;
+}
+
+/** 
+* @minimum 0
+* @maximum 18446744073709551615
+* @asType integer
+*/
+export type u64 = number;
+/** 
+* @minimum -9223372036854775808
+* @maximum 9223372036854775807
+* @asType integer
+*/
+export type i64 = number;
+
+/**
+* @minimum  0 
+* @maximum 255
+* @asType integer
+* */
+export type u8 = number;
+/**
+* @minimum  -128 
+* @maximum 127
+* @asType integer
+* */
+export type i8 = number;
+/**
+* @minimum  0 
+* @maximum 65535
+* @asType integer
+* */
+export type u16 = number;
+/**
+* @minimum -32768 
+* @maximum 32767
+* @asType integer
+* */
+export type i16 = number;
+/**
+* @minimum 0 
+* @maximum 4294967295
+* @asType integer
+* */
+export type u32 = number;
+/**
+* @minimum 0 
+* @maximum 4294967295
+* @asType integer
+* */
+export type usize = number;
+/**
+* @minimum  -2147483648 
+* @maximum 2147483647
+* @asType integer
+* */
+export type i32 = number;
+
+/**
+* @minimum -3.40282347E+38
+* @maximum 3.40282347E+38
+*/
+export type f32 = number;
+
+/**
+* @minimum -1.7976931348623157E+308
+* @maximum 1.7976931348623157E+308
+*/
+export type f64 = number;";
